import math
import random
from copy import deepcopy

import numpy as np
import pytest
import torch
from tokenizers import Tokenizer, models
from transformers import AutoTokenizer, PreTrainedTokenizerFast

from ntloss import NTLoss, NTLossDotProduct, NumberLevelLoss
from ntloss.utils import is_number

TOKENIZER = AutoTokenizer.from_pretrained("t5-small")
VOCAB_SIZE = TOKENIZER.vocab_size


def get_device(use_cpu: bool = False) -> str:
    """Get available device.

    Returns:
        Device
    """
    if not use_cpu:
        if torch.cuda.is_available():
            return "cuda"
        elif torch.backends.mps.is_available() and torch.backends.mps.is_built():
            return "mps"
    return "cpu"


DEVICE = get_device()


def make_logits(token_logit_value_dicts):
    """
    Build a (1 x T x V) tensor filled with -inf,
    then set the logits specified in token_logit_value_dicts.
    """
    seq_len = len(token_logit_value_dicts)
    logits = torch.full((1, seq_len, VOCAB_SIZE), -np.inf, dtype=torch.float32)
    for i, tok_dict in enumerate(token_logit_value_dicts):
        for tok_id, logit in tok_dict.items():
            logits[0, i, tok_id] = logit
    return logits.to(device=DEVICE)


def dirac_logits(ids, peak_id, peak_value, vocab_size=VOCAB_SIZE):
    """Perfectly confident distribution (all mass on one token)."""
    logits = torch.full((1, 1, vocab_size), -1e6, dtype=torch.float32)
    logits[0, 0, peak_id] = 0.0
    return logits.to(device=DEVICE)


def gaussian_logits(ids, peak_id, peak_value, vocab_size=VOCAB_SIZE, sigma=1e-1):
    """
    Smooth bell curve over the reference tokens.
    """
    logits = torch.full((1, 1, vocab_size), -1e6, dtype=torch.float32)
    # Assumes that ids are ordered by their numerical value
    for idx, tok_id in enumerate(ids):
        logits[0, 0, tok_id] = -1 * np.abs((idx - peak_value) ** 2 / (2 * sigma**2))
    return logits.to(device=DEVICE)


@pytest.mark.parametrize("loss_class", [NTLoss, NTLossDotProduct])
@pytest.mark.parametrize(
    "logits_dicts,label_tokens",
    [
        # positive logits scenario
        (
            [
                {"1": 1.0, "2": 1.2, "0": 0.5, "3": 1.5},
                {"1": 1.5, "2": 1.2, "0": 0.5, "3": 1.5},
                {"1": 1.0, "2": 1.2, "0": 0.5, "3": 1.5},
            ],
            ["1", "1", "a"],
        ),
        # mixed logits scenario
        (
            [
                {"0": -4.0, "1": 2.0, "2": -1.0},
                {"0": 1.5, "1": 0.5, "2": 1.2},
                {"3": -2.0, "4": 1.0, "5": -2.5},
            ],
            ["2", "1", "3"],
        ),
    ],
)
@pytest.mark.parametrize("reduction", ["mean", "sum", "none"])
@pytest.mark.parametrize("loss_weight", [1, 0, 0.5, None])
def test_ntloss_variants(
    loss_class, logits_dicts, label_tokens, reduction, loss_weight
):
    # convert token strings to IDs
    token_logit_value_dicts = {
        # map token strings to token IDs upfront
        i: {
            TOKENIZER.convert_tokens_to_ids(tok): val
            for tok, val in logits_dicts[i].items()
        }
        for i in range(len(logits_dicts))
    }
    # build logits tensor
    # reorder into a list for our helper
    logits_list = [token_logit_value_dicts[i] for i in range(len(logits_dicts))]
    logits = make_logits(logits_list)

    # build labels tensor shape (1 x T)
    label_ids = [TOKENIZER.convert_tokens_to_ids(tok) for tok in label_tokens]
    labels = torch.tensor([label_ids], dtype=torch.long, device=logits.device)

    # set up loss weights
    if loss_weight is None:
        loss_weights = None
    else:
        loss_weights = torch.ones_like(labels) * loss_weight

    # instantiate and run
    loss_fn = loss_class(tokenizer=TOKENIZER)
    loss = loss_fn(logits, labels, reduction=reduction, loss_weights=loss_weights)

    assert torch.is_floating_point(loss), "Loss should be a floating tensor"
    if reduction != "none":
        assert isinstance(loss.item(), float), "Loss should be a Python float"
        assert not math.isnan(loss), "Loss must not be NaN"
        assert loss.item() > 0 or loss_weights.sum() == 0, "Loss must be positive"
    else:
        assert loss.shape == labels.shape, (
            "Loss and labels must have same shape if reduction is none"
        )
        assert isinstance(loss.sum().item(), float), "Loss should be a Python float"
        assert not math.isnan(loss.sum()), "Loss must not be NaN"
        assert loss.sum().item() > 0 or loss_weights.sum() == 0, "Loss must be positive"


@pytest.mark.parametrize("loss_class", [NTLoss, NTLossDotProduct])
@pytest.mark.parametrize("reweigh", [True, False])
def test_differentiability(loss_class, reweigh):
    loss_fn = loss_class(TOKENIZER, reweigh=reweigh)

    ref_tokens = [str(i) for i in range(10)] + ["A"]
    ref_ids = [TOKENIZER.convert_tokens_to_ids(t) for t in ref_tokens]

    gt_token_id = ref_ids[1]
    p_token_id = ref_ids[5]
    labels = torch.tensor([[gt_token_id]], dtype=torch.long)

    logits = torch.full(
        (1, 1, VOCAB_SIZE), -1e6, dtype=torch.float32, requires_grad=True
    )
    logits_ = torch.full((1, 1, VOCAB_SIZE), 0.0, dtype=torch.float32)
    logits_[:, :, p_token_id] = 1e6
    logits = logits + logits_

    loss = loss_fn(logits, labels)

    assert loss.grad_fn is not None, "Loss is not differentiable!"


@pytest.mark.parametrize("loss_class", [NTLoss, NTLossDotProduct])
@pytest.mark.parametrize("logit_builder", [dirac_logits, gaussian_logits])
def test_correct_minimum(loss_class, logit_builder):
    loss_fn = loss_class(TOKENIZER, reweigh=True)
    ref_tokens = [str(i) for i in range(10)] + ["A"]
    ref_ids = [TOKENIZER.convert_tokens_to_ids(t) for t in ref_tokens]

    # Guard: make sure all required tokens exist in the vocab
    assert all(i is not None and i >= 0 for i in ref_ids), "Missing token id"

    losses = torch.zeros(len(ref_ids), len(ref_ids), dtype=torch.float32)
    for i, (gt_token, gt_token_id) in enumerate(zip(ref_tokens, ref_ids)):
        labels = torch.tensor([[gt_token_id]], dtype=torch.long)
        for peak_idx, peak_id in enumerate(ref_ids):
            logits = logit_builder(ref_ids, peak_id, peak_idx)
            loss = loss_fn(logits, labels.to(device=logits.device))

            losses[i, peak_idx] = loss.item()

            if isinstance(loss_fn, NTLossDotProduct):
                yhat, nt_mass = loss_fn.predict_numbers(logits)
                assert 0 <= nt_mass <= 1
                assert 0 <= yhat <= 9

        # TODO: Ensure that if GT is number and mass is on text, loss is at least as
        # high as for worst number prediction. This is not there yet and the reason
        # why we exclude the last token
        mins = torch.argsort(losses[i, :-1], dim=0)
        expected = torch.Tensor(
            sorted(range(10), key=lambda j: (abs(j - i), j)),
        ).long()

        if i == 10:
            assert torch.allclose(
                losses[i, :], torch.zeros_like(losses[i, :]), atol=1e-8
            ), "Loss should be zero when the ground-truth token is non-numeric."
        else:
            assert torch.equal(mins, expected), (
                "For a digit ground truth, loss must be minimal when the distribution "
                "peaks over the same digit."
            )

    assert not torch.isnan(losses).any(), "Encountered NaN in loss matrix"


@pytest.mark.parametrize(
    "custom_vocab_li",
    [
        None,
        random.shuffle(list((range(0, 10, 2)))),
        random.shuffle(list(range(0, 10, 2)) + [100]),
        random.shuffle(list(range(0, 100, 10))),
    ],
)
@pytest.mark.parametrize("loss_class", [NTLoss])
@pytest.mark.parametrize("logit_builder", [dirac_logits, gaussian_logits])
@pytest.mark.parametrize("squash_factor", [0.5, 1, 2, 20])
def test_setup_distance_lookup(
    custom_vocab_li, loss_class, logit_builder, squash_factor
):
    # Make sure mapping of order of NTs in vocab is maintained for dist_matrix
    # Make sure that the distance matrix doesn't make assumptions about order of NTs

    if custom_vocab_li is not None:
        nums_in_vocab = custom_vocab_li
        custom_vocab = dict(
            [(str(n), i) for i, n in enumerate(random.shuffle(list((range(0, 10, 2)))))]
        )
        if "A" not in custom_vocab:
            custom_vocab["A"] = len(custom_vocab)
        tok = Tokenizer(models.WordLevel(vocab=custom_vocab))
        tokenizer = PreTrainedTokenizerFast(tokenizer_object=tok)
    else:
        nums_in_vocab = list(range(10))
        tokenizer = TOKENIZER

    loss_fn = loss_class(tokenizer, digit_level=False)

    num_vals = loss_fn.number_values_dense

    assert loss_fn.dist_lookup.shape[0] == loss_fn.dist_lookup.shape[1], (
        "The distance lookup matrix should be square."
    )
    assert loss_fn.dist_lookup.shape[0] == len(num_vals), (
        "The distance lookup matrix should contain distances for all number tokens "
        "in vocab."
    )

    # Check whether value at (i,j) in matrix is the expected value: the absolute
    # difference between num_vals[i] and num_vals[j]
    for i, i_val in enumerate(num_vals):
        for j, j_val in enumerate(num_vals):
            assert loss_fn.dist_lookup[i, j] == abs(i_val - j_val), (
                f"Value in cell {i}, {j} of the distance lookup matrix is not the "
                f"expected value ({loss_fn.dist_lookup[i, j]} vs {abs(i_val - j_val)})."
            )

    ref_tokens = [str(i) for i in nums_in_vocab]
    ref_ids = [TOKENIZER.convert_tokens_to_ids(t) for t in ref_tokens]

    # Check whether all num_ids have a mapping from idx in vocab to idx in dist lookup
    num_ids = torch.where(loss_fn.vocab_to_dist_idx != -1)[0]
    for i in ref_ids:
        assert i in num_ids, f"Couldn't find {i} in vocab_to_dist_idx mapping."

    # Check whether mapping from vocab to the index in dist lookup matrix is correct
    for i, ref_idx in enumerate(ref_ids):
        dist_idx = loss_fn.vocab_to_dist_idx[ref_idx]

        for j, j_val in enumerate(num_vals):
            i_val = float(ref_tokens[i])
            assert loss_fn.dist_lookup[dist_idx, j] == abs(i_val - j_val), (
                f"Value in cell {dist_idx}, {j} of the distance lookup matrix is not the "
                f"expected value ({loss_fn.dist_lookup[dist_idx, j]} vs {abs(i_val - j_val)})."
            )


@pytest.mark.parametrize("loss_class", [NTLoss])
@pytest.mark.parametrize("logit_builder", [dirac_logits, gaussian_logits])
@pytest.mark.parametrize("squash_factor", [0.5, 1, 2, 20])
def test_correct_squashing(loss_class, logit_builder, squash_factor):
    # Make sure wrong uses of the squashing factor are caught
    if squash_factor <= 1:
        with pytest.raises(
            AssertionError,
            match=r"The squash factor can't be equal to or smaller than 1*",
        ):
            loss_fn = loss_class(TOKENIZER, squash_factor=squash_factor)

        return

    loss_fn = loss_class(TOKENIZER, squash_factor=squash_factor, reweigh=False)
    ref_tokens = [str(i) for i in range(10)] + ["A"]
    ref_ids = [TOKENIZER.convert_tokens_to_ids(t) for t in ref_tokens]

    # Guard: make sure all required tokens exist in the vocab
    assert all(i is not None and i >= 0 for i in ref_ids), "Missing token id"

    # Make sure that the maximum distance in loss' distance matrix is larger than
    # nonzero minimum, with a factor equal to the squash_factor used
    inf = torch.finfo(loss_fn.dist_lookup.dtype).max
    lookup_nz = loss_fn.dist_lookup.masked_fill(loss_fn.dist_lookup == 0, inf)
    min_lookup_nz = lookup_nz.min()
    max_lookup = loss_fn.dist_lookup.max()

    computed = torch.div(max_lookup, min_lookup_nz)
    expected = torch.tensor([squash_factor], dtype=loss_fn.dist_lookup.dtype)
    assert torch.allclose(computed, expected, atol=1e-8), (
        "Distance to farthest number token should be the defined squashing factor larger ",
        f"than the distance to the closest number token ({computed} instead of {expected}).",
    )

    # Also make sure that the number token loss is thus always smaller or equal to the squashing factor
    losses = torch.zeros(len(ref_ids), len(ref_ids), dtype=torch.float32)
    for i, (gt_token, gt_token_id) in enumerate(zip(ref_tokens, ref_ids)):
        labels = torch.tensor([[gt_token_id]], dtype=torch.long)
        for peak_idx, peak_id in enumerate(ref_ids):
            logits = logit_builder(ref_ids, peak_id, peak_idx)
            loss = loss_fn(logits, labels.to(device=logits.device))
            losses[i, peak_idx] = loss.item()

    assert not torch.isnan(losses).any(), "Encountered NaN in loss matrix"

    assert torch.all(losses <= squash_factor), (
        "Loss should be smaller or equal to the squashing factor, if this is set."
    )


@pytest.mark.parametrize(
    "custom_vocab",
    [
        None,
        dict([(str(n), i) for i, n in enumerate(range(0, 10, 2))]),
        dict([(str(n), i) for i, n in enumerate(list(range(0, 10, 2)) + [100])]),
        dict([(str(n), i) for i, n in enumerate(range(0, 100, 10))]),
    ],
)
@pytest.mark.parametrize("loss_class", [NTLoss])
@pytest.mark.parametrize("logit_builder", [dirac_logits, gaussian_logits])
@pytest.mark.parametrize("squash_factor", [None, 0.5, 1, 2, 20])
def test_irregular_nt_vocab(custom_vocab, loss_class, logit_builder, squash_factor):
    if custom_vocab is not None:
        if "A" not in custom_vocab:
            custom_vocab["A"] = len(custom_vocab)
        tok = Tokenizer(models.WordLevel(vocab=custom_vocab))
        tokenizer = PreTrainedTokenizerFast(tokenizer_object=tok)
    else:
        tokenizer = TOKENIZER
    vocab_size = tokenizer.vocab_size

    # Make sure wrong uses of the squashing factor are caught
    if squash_factor is not None and squash_factor <= 1:
        with pytest.raises(
            AssertionError,
            match=r"The squash factor can't be equal to or smaller than 1*",
        ):
            loss_fn = loss_class(
                tokenizer,
                digit_level=False,
                squash_factor=squash_factor,
            )

        return

    loss_fn = loss_class(
        tokenizer,
        digit_level=False,
        squash_factor=squash_factor,
        reweigh=False,
    )
    ref_tokens = [str(i) for i in range(10)] + ["A"]
    ref_ids = [tokenizer.convert_tokens_to_ids(t) for t in ref_tokens]

    # Remove tokens not present in vocab from ref_tokens and ref_ids lists
    ref_tokens = [tok for i, tok in enumerate(ref_tokens) if ref_ids[i] is not None]
    ref_ids = [i for i in ref_ids if i is not None]

    # Make sure that the maximum distance in loss' distance matrix is larger than
    # nonzero minimum, with a factor equal to the squash_factor used
    inf = torch.finfo(loss_fn.dist_lookup.dtype).max
    lookup_nz = loss_fn.dist_lookup.masked_fill(loss_fn.dist_lookup == 0, inf)
    min_lookup_nz = lookup_nz.min()
    max_lookup = loss_fn.dist_lookup.max()

    computed = torch.div(max_lookup, min_lookup_nz)

    if custom_vocab is not None:
        max_nt = max([float(t) for t in custom_vocab.keys() if is_number(t)])
        min_nt = min(
            [float(t) for t in custom_vocab.keys() if is_number(t) and float(t) > 0]
        )
        factor = squash_factor or max_nt / min_nt
        expected = torch.tensor([factor], dtype=loss_fn.dist_lookup.dtype)
        assert torch.allclose(computed, expected, atol=1e-8), (
            "Distance to farthest number token should be the defined squashing factor larger ",
            f"than the distance to the closest number token ({computed} instead of {expected}).",
        )

    # Also make sure that the number token loss is thus always smaller or equal to the squashing factor
    losses = torch.zeros(len(ref_ids), len(ref_ids), dtype=torch.float32)
    for i, (gt_token, gt_token_id) in enumerate(zip(ref_tokens, ref_ids)):
        labels = torch.tensor([[gt_token_id]], dtype=torch.long)
        for peak_idx, peak_id in enumerate(ref_ids):
            logits = logit_builder(ref_ids, peak_id, peak_idx, vocab_size)
            loss = loss_fn(logits, labels.to(device=logits.device))
            losses[i, peak_idx] = loss.item()

    assert not torch.isnan(losses).any(), "Encountered NaN in loss matrix"
    if squash_factor is not None:
        assert torch.all(losses <= squash_factor), (
            "Loss should be smaller or equal to the squashing factor, if this is set."
        )


@pytest.mark.parametrize("loss_class", [NTLoss, NTLossDotProduct])
@pytest.mark.parametrize("logit_builder", [dirac_logits, gaussian_logits])
def test_logit_scaling(loss_class, logit_builder):
    loss_fn = loss_class(TOKENIZER, reweigh=True)
    ref_tokens = [str(i) for i in range(10)] + ["A"]
    ref_ids = [TOKENIZER.convert_tokens_to_ids(t) for t in ref_tokens]

    # Make sure max_dist is set up correctly for regularization
    assert loss_fn.max_dist > 0, "loss_fn.max_dist is not set up correctly"

    # Guard: make sure all required tokens exist in the vocab
    assert all(i is not None and i >= 0 for i in ref_ids), "Missing token id"

    losses = torch.zeros(len(ref_ids), len(ref_ids), dtype=torch.float32)
    for i, (gt_token, gt_token_id) in enumerate(zip(ref_tokens, ref_ids)):
        labels = torch.tensor([[gt_token_id]], dtype=torch.long)
        for peak_idx, peak_id in enumerate(ref_ids):
            logits = logit_builder(ref_ids, peak_id, peak_idx)
            loss = loss_fn(logits, labels.to(device=logits.device))
            losses[i, peak_idx] = loss.item()

        # Ensure that if GT is number and mass is on text, loss is at least as
        # high as for worst number prediction: should be the case for weighted NT loss.
        mins = torch.argsort(losses[i, :], dim=0)
        expected = torch.Tensor(
            sorted(range(10), key=lambda j: (abs(j - i), j)) + [10],
        ).long()

        if i == 10:
            assert torch.allclose(
                losses[i, :], torch.zeros_like(losses[i, :]), atol=1e-8
            ), "Loss should be zero when the ground-truth token is non-numeric."
        else:
            assert torch.equal(mins, expected), (
                "For a digit ground truth, loss must be minimal when the distribution "
                "peaks over the same digit."
            )
            assert torch.all(losses[i, expected[-2]] <= losses[i, -1]), (
                "For a digit ground truth and mass concentration on text, weighted loss "
                "should be at least as high as for worst number prediction."
            )

    assert not torch.isnan(losses).any(), "Encountered NaN in loss matrix"


def test_digit_level():
    # Ensure that 10 Number tokens are extracted if digit_level is set to True
    NEW_TOKENIZER = deepcopy(TOKENIZER)
    loss_class = NTLoss(tokenizer=NEW_TOKENIZER, digit_level=True)
    assert len(loss_class.number_values_dense) == 10

    # Add some tokens that are in right range but should still be ignored
    NEW_TOKENIZER.add_tokens([" 2"])
    loss_class = NTLoss(tokenizer=NEW_TOKENIZER, digit_level=True)
    assert len(loss_class.number_values_dense) == 10


def test_number_level_ntl():
    seq_tokens = ["A", "1", "2", "3", "B", "4", "5"]
    label_ids = TOKENIZER.convert_tokens_to_ids(seq_tokens)
    assert all(x is not None and x >= 0 for x in label_ids), "Missing token id"
    labels = torch.tensor([label_ids], dtype=torch.long)

    def one_hot_pos(tok_id, logit=50.0):
        return {tok_id: logit}

    # Perfect prediction: put all mass on the correct digit at each digit position.
    logits_dicts_perfect = []
    for tok in seq_tokens:
        tid = TOKENIZER.convert_tokens_to_ids(tok)
        # for non-digits, it doesn't matter; keep mass on that token to be safe
        logits_dicts_perfect.append(one_hot_pos(tid, 50.0))
    logits_perfect = make_logits(logits_dicts_perfect)
    labels = labels.to(device=logits_perfect.device)

    loss_fn = NumberLevelLoss(TOKENIZER, reweigh=False)
    loss = loss_fn(logits_perfect, labels, reduction="none")
    assert loss.shape == labels.shape
    assert loss.sum().item() == 0

    # Sanity: perfect case should be exactly zero for mean/sum as well
    loss = loss_fn(logits_perfect, labels, reduction="mean")
    assert loss.sum().item() == 0
    loss = loss_fn(logits_perfect, labels, reduction="sum")
    assert loss.sum().item() == 0

    # Now make a single digit wrong: change the middle digit "2" -> predict "9" instead.
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[2] = {TOKENIZER.convert_tokens_to_ids("9"): 50.0}
    logits_middle_wrong = make_logits(wrong_logits_dicts)

    loss_middle = loss_fn(logits_middle_wrong, labels, reduction="none").squeeze()
    # Check that loss for that item is nonzero
    assert loss_middle[1].item() > 0
    # All others should have zero loss
    assert loss_middle[0].item() == 0 and loss_middle[2:].sum() == 0

    # Now make the first digit wrong and check whether error is higher
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[1] = {TOKENIZER.convert_tokens_to_ids("9"): 50.0}
    logits_first_wrong = make_logits(wrong_logits_dicts)

    loss_first = loss_fn(logits_first_wrong, labels, reduction="none").squeeze()
    assert loss_first[1].item() > 0
    assert loss_first[0].item() == 0 and loss_first[2:].sum() == 0
    assert loss_first[1].item() > loss_middle[1].item()

    # Now make the third digit wrong and check whether error is lower
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[3] = {TOKENIZER.convert_tokens_to_ids("5"): 50.0}
    logits_last_wrong = make_logits(wrong_logits_dicts)

    loss_last = loss_fn(logits_last_wrong, labels, reduction="none").squeeze()
    assert loss_last[1].item() > 0
    assert loss_last[0].item() == 0 and loss_last[2:].sum() == 0
    assert loss_last[1].item() < loss_middle[1].item()

    # If we aggregate, this must still hold
    for reduction in ["mean", "sum"]:
        loss_middle = loss_fn(
            logits_middle_wrong, labels, reduction=reduction
        ).squeeze()
        loss_last = loss_fn(logits_last_wrong, labels, reduction=reduction).squeeze()
        loss_first = loss_fn(logits_first_wrong, labels, reduction=reduction).squeeze()

        assert loss_middle.item() > 0.0
        assert loss_last.item() > 0.0
        assert loss_first.item() > 0.0
        assert loss_first.item() > loss_middle.item() > loss_last.item()


@pytest.mark.parametrize("reweigh", [True, False])
def test_number_level_ntl_scientific_notation(reweigh: bool):
    seq_tokens = ["A", "6", ".", "5", "1", "E", "+", "0", "2"]
    label_ids = TOKENIZER.convert_tokens_to_ids(seq_tokens)
    assert all(x is not None and x >= 0 for x in label_ids), "Missing token id"
    labels = torch.tensor([label_ids], dtype=torch.long)

    def one_hot_pos(tok_id, logit=50.0):
        return {tok_id: logit}

    # Perfect prediction: put all mass on the correct digit at each digit position.
    logits_dicts_perfect = []
    for tok in seq_tokens:
        tid = TOKENIZER.convert_tokens_to_ids(tok)
        # for non-digits, it doesn't matter; keep mass on that token to be safe
        logits_dicts_perfect.append(one_hot_pos(tid, 50.0))
    logits_perfect = make_logits(logits_dicts_perfect)
    logits_perfect.requires_grad = True
    labels = labels.to(device=logits_perfect.device)

    loss_fn = NumberLevelLoss(TOKENIZER, reweigh=reweigh, float_level=True)
    loss = loss_fn(logits_perfect, labels, reduction="none")
    assert loss.shape == labels.shape
    assert loss.sum().item() == 0
    assert loss.grad_fn is not None, "Loss is not differentiable!"

    # Sanity: perfect case should be exactly zero for mean/sum as well
    loss = loss_fn(logits_perfect, labels, reduction="mean")
    assert loss.sum().item() == 0
    loss = loss_fn(logits_perfect, labels, reduction="sum")
    assert loss.sum().item() == 0

    # Now make a single digit in 6.51 wrong: change the middle digit "5" -> predict "4" instead.
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[3] = {TOKENIZER.convert_tokens_to_ids("6"): 50.0}
    logits_middle_wrong = make_logits(wrong_logits_dicts)

    loss_middle = loss_fn(logits_middle_wrong, labels, reduction="none").squeeze()
    # Check that loss for that item is nonzero
    assert loss_middle[1].item() > 0
    # All others should have zero loss
    assert loss_middle[0].item() == 0 and loss_middle[2:].sum() == 0

    # Now make the first digit wrong and check whether error is higher
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[1] = {TOKENIZER.convert_tokens_to_ids("7"): 50.0}
    logits_first_wrong = make_logits(wrong_logits_dicts)

    loss_first = loss_fn(logits_first_wrong, labels, reduction="none").squeeze()
    assert loss_first[1].item() > 0
    assert loss_first[0].item() == 0 and loss_first[2:].sum() == 0
    assert loss_first[1].item() > loss_middle[1].item()

    # Now make the third digit wrong and check whether error is lower
    wrong_logits_dicts = [d.copy() for d in logits_dicts_perfect]
    wrong_logits_dicts[4] = {TOKENIZER.convert_tokens_to_ids("2"): 50.0}
    logits_last_wrong = make_logits(wrong_logits_dicts)

    loss_last = loss_fn(logits_last_wrong, labels, reduction="none").squeeze()
    assert loss_last[1].item() > 0
    assert loss_last[0].item() == 0 and loss_last[2:].sum() == 0
    assert loss_last[1].item() < loss_middle[1].item()

    # Test loss weights
    loss_weights = torch.rand_like(labels, dtype=logits_last_wrong.dtype)
    loss_with_weights = loss_fn(
        logits_last_wrong, labels, loss_weights=loss_weights, reduction="none"
    ).squeeze()
    assert loss_with_weights[1].item() > 0
    assert loss_with_weights[0].item() == 0 and loss_with_weights[2:].sum() == 0
    assert loss_with_weights[1].item() < loss_last[1].item()

    # If we aggregate, this must still hold
    for reduction in ["mean", "sum"]:
        loss_middle = loss_fn(
            logits_middle_wrong, labels, reduction=reduction
        ).squeeze()
        loss_last = loss_fn(logits_last_wrong, labels, reduction=reduction).squeeze()
        loss_first = loss_fn(logits_first_wrong, labels, reduction=reduction).squeeze()

        assert loss_middle.item() > 0.0
        assert loss_last.item() > 0.0
        assert loss_first.item() > 0.0
        assert loss_first.item() > loss_middle.item() > loss_last.item()



from abc import ABC, abstractmethod
from numbers import Number
from typing import Callable, Optional, Tuple, cast

import torch
import torch.nn.functional as F
from loguru import logger
from torch import BoolTensor, FloatTensor, LongTensor, Tensor
from transformers import PreTrainedTokenizer

from .utils import is_number


class AbstractNTLoss(ABC):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        digit_level: bool = True,
        reweigh: bool = True,
    ):
        """
        NTL constructor.

        Args:
            tokenizer: Standard HF tokenizer.
            digit_level: Whether to ensure only digits are considered number tokens,
                stabilizing training with NTL. Defaults to True. Used for most
                experiments in the ICML paper.
            reweigh: Whether to scale the NTL using the logit weight on
                number tokens. Defaults to True.
                NOTE: The ICML paper does *not* use this option which can lead to
                incorrect loss if most mass is placed outside of the number tokens.

        """
        super().__init__()
        self.tokenizer = tokenizer
        self.digit_level = digit_level
        self.reweigh = reweigh

        self.setup_number_tokens()

        self.max_dist = torch.tensor(0.0)

    def setup_number_tokens(self):
        """Setting up attributes needed by NT loss"""

        # Add digits to vocab if not there yet.
        vocab_size = len(self.tokenizer)
        if self.digit_level:
            new_tokens = self.tokenizer.add_tokens(list(map(str, range(10))))
        if vocab_size < len(self.tokenizer) and new_tokens > 0:
            logger.warning(f"Added {new_tokens} new tokens for number token loss")
        vocab = self.tokenizer.get_vocab()
        self.number_values: FloatTensor = torch.full((len(vocab),), float("nan"))

        # Try to convert each token to a float after stripping the space prefix
        for token, id in vocab.items():
            if is_number(token, finite=True):
                if self.digit_level:
                    # NOTE: This check ensures number token value only occurs for digits, not for multi-digit numbers (123)
                    # This stabilizes training with NTL. Can be altered though, see paper experiments.
                    # Excludes tokens that are numbers in other languages like ႘ and tokens with space pre-/postfix like ` 2`.
                    if token.isascii() and -1 <= float(token) <= 9 and len(token) == 1:
                        self.number_values[id] = float(token)
                else:
                    self.number_values[id] = float(token)

        self.is_number_token = ~torch.isnan(self.number_values)
        if self.is_number_token.sum() == len(self.is_number_token):
            raise ValueError(
                "At least one token needs to be not a number, otherwise `ignore_index` cannot be set up safely"
            )
        self.nan_id = torch.where(~self.is_number_token)[0][0].item()
        self.number_values_dense = self.number_values[self.is_number_token]

        if self.digit_level and (num_nts := len(self.number_values_dense)) != 10:
            logger.error(
                f"You requested digit-level but {num_nts} number tokens were identified: {self.number_values_dense}"
            )

    @abstractmethod
    def forward(
        self,
        logits: FloatTensor,
        labels: LongTensor,
        loss_weights: Optional[Tensor] = None,
        reduction: str = "mean",
    ) -> Tensor: ...

    def __call__(self, *args, **kwargs):
        """Alias to self.forward"""
        return self.forward(*args, **kwargs)

    def reweigh_fn(
        self,
        logits: Tensor,
        loss: Tensor,
        number_token_positions: Tensor,
    ) -> Tensor:
        """
        Scale the NT loss element-wise using the logit weight on number tokens.
        NOTE: This reweighing ensures that if ground truth is a number token
            but most probability mass is on text tokens, the loss will be *higher*
            than the worst possible number token. This is an edge case in practice.

        Args:
            logits: 3D Tensor of shape BS x T x V.
            loss: 1D Tensor over all number tokens in batch.
            number_token_positions: 2D Tensor of shape BS x T indicating for which tokens
                the NT loss was computed.

        Returns:
            A 1D Tensor over all number tokens in batch with the scaled NT losses.
        """

        # Take softmax over logits of all tokens in vocab and compute NT logit weight
        softmax_probs_all = F.softmax(logits, dim=-1)
        nt_logit_weight = torch.sum(
            softmax_probs_all[:, :, self.is_number_token], dim=-1
        )[number_token_positions]

        # Apply weights for NTL element-wise
        loss *= nt_logit_weight

        # Apply regularization
        # NOTE: We could consider reweighing here with the max for that label token
        # rather than the global max
        loss += (
            1.01
            * self.max_dist.to(dtype=loss.dtype, device=loss.device)
            * (1 - nt_logit_weight)
        )

        return loss

    @staticmethod
    def _validate_inputs(
        logits: FloatTensor,
        labels: Optional[LongTensor],
        loss_weights: Optional[Tensor],
    ):
        """Private method to perform size and type checks."""
        if (td := len(logits.shape)) != 3 or (ne := logits.numel()) == 0:
            raise ValueError(
                f"Logits have to be non-empty 3D Tensor, not {td}D with {ne} elements"
            )
        if not torch.is_floating_point(logits):
            raise TypeError("Logits have to be FloatTensor.")
        if labels is None:
            return
        if not labels.dtype == torch.long:
            raise TypeError(f"Labels have to be LongTensor, not {type(labels)}")
        if (b := labels.shape) != (a := logits.shape[:-1]):
            raise ValueError(
                f"Logit and label sizes of first 2 dims have to match: {a} vs {b}"
            )

        if (td := len(labels.shape)) != 2 or (ne := labels.numel()) == 0:
            raise ValueError(
                f"Labels have to be non-empty 2D Tensor, not {td}D with {ne} elements"
            )
        if loss_weights is not None:
            if loss_weights.shape != labels.shape:
                raise ValueError(
                    "Loss mask has to be 2D Tensor of same shape as labels."
                )
            if torch.any(loss_weights < 0):
                raise ValueError("loss_mask must be ≥ 0.")

    def _prepare_number_token_targets(
        self, labels: LongTensor, loss_weights: Optional[Tensor], ignore_index: int
    ) -> Tuple[FloatTensor, Tensor]:
        """
        Prepare number-token targets and masks.

        Args:
            labels: 2D Tensor of shape BS x T.
            loss_weights: Optional 2D Tensor of shape BS x T with loss weight for each token.
            ignore_index: Label ID to ignore. Defaults to -100.

        Returns:
            y: 2D Float Tensor of shape BS x T with target numeric values (NaN for non-number tokens).
            loss_weight: 1D Tensor with a potentially individual loss weight for each number token position.
        """
        labels = cast(
            LongTensor, labels.masked_fill(labels == ignore_index, self.nan_id)
        )
        # Create a mask to filter out non-digit tokens
        y = self.number_values.to(device=labels.device)[labels]
        number_token_positions = ~torch.isnan(y)
        loss_weights = (
            loss_weights[number_token_positions]
            if loss_weights is not None
            else torch.ones_like(labels, device=labels.device)[number_token_positions]
        )
        return cast(FloatTensor, y), loss_weights

    @staticmethod
    def _apply_reduction(
        loss: Tensor,
        reduction: str,
        loss_weights: Tensor,
        number_token_positions: Tensor,
        logits: Tensor,
    ) -> Tensor:
        """
        Applies the specified reduction type to the calculated loss.

        This method handles 3 types of reduction: "mean", "sum", and "none".
        For "mean" and "sum", it applies weighting using `loss_weights`.
        For "none", it reshapes the loss back to the original batch and sequence
        dimensions.

        Args:
            loss: 1D Tensor containing the loss for each number token in the batch.
            reduction: The reduction method ("mean", "sum", or "none").
            loss_weights: 1D Tensor with a loss weight for each number token.
            number_token_positions: 2D boolean tensor of shape BS x T indicating
                the positions of number tokens.
            logits: 3D Tensor of shape BS x T x V, used to get the original shape
                for the "none" reduction.

        Returns:
            A Tensor representing the reduced loss:
                - 0D tensor if `reduction` is "mean" or "sum".
                - 2D Tensor of shape BS x T if `reduction` is "none".
        """
        if reduction == "mean":
            # Mean pooling (weighted by loss mask)
            loss = torch.dot(
                loss.flatten(), loss_weights.flatten()
            ) / loss_weights.sum().clamp_min(torch.finfo(loss.dtype).eps)
        elif reduction == "sum":
            loss = torch.dot(loss.flatten(), loss_weights.flatten())
        elif reduction == "none":
            # Cast loss for number tokens back to Tensor of size BS x T
            loss_ = torch.zeros(number_token_positions.numel()).to(loss.device)
            loss_[number_token_positions.view(-1)] = loss * loss_weights
            bs, seq_len, _ = logits.size()
            loss = loss_.view(bs, seq_len)

            assert torch.sum(loss[~number_token_positions]) == 0, (
                "NumberTokenLoss computed for non-digit tokens!"
            )

        else:
            raise ValueError(f"{reduction} is not a valid value for reduction")

        return loss


class NTLossDotProduct(AbstractNTLoss):
    """Class for NT losses that produce a token-wise numerical output."""

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        digit_level: bool = True,
        reweigh: bool = True,
        loss_function: Callable = F.mse_loss,
    ):
        """
        Referred to as NTL-L_p in the paper.

        Args:
            tokenizer: NTLTokenizer with necessary attributes like is_number_token etc.
            digit_level: Whether to ensure only digits are considered number tokens,
                stabilizing training with NTL. Defaults to True. Used for most
                experiments in the ICML paper.
            reweigh: Whether to scale the NTL using the logit weight on
                number tokens. Defaults to True.
                NOTE: The ICML paper does *not* use this option which can lead to
                incorrect loss if most mass is placed outside of the number tokens.
            loss_function: Function to apply on the delta between the ground truth number
                and the obtained dot product (nt-probs * token-values). Defaults to
                MSE, but MAE, Huber etc are also compatible.
        """
        super().__init__(
            tokenizer=tokenizer,
            digit_level=digit_level,
            reweigh=reweigh,
        )
        self.loss_function = loss_function
        self.setup_max_dist()

    def setup_max_dist(self):
        """
        Set up the maximum distance between the number tokens based on the selected loss function.
        """

        # Extract the number token values and get the minimum and maximum
        vals = self.number_values_dense.unsqueeze(0)
        max_val = vals.max()
        min_val = vals.min()

        # Compute the largest value the loss function used in NT loss computation can get
        # Make sure to account for possibility of asymmetrical loss function
        self.max_dist = torch.maximum(
            torch.abs(self.loss_function(min_val, max_val)),
            torch.abs(self.loss_function(max_val, min_val)),
        )

    def predict_numbers(self, logits: FloatTensor) -> Tuple[FloatTensor, FloatTensor]:
        """
        Calculates token-level numerical prediction.
        NOTE: This calculates numerical predictions for *all* tokens, not just where
        label is a number token.

        Args:
            logits: 3D FloatTensor of shape BS x T x V.

        Returns:
            yhat: 2D FloatTensor BS x T containing numerical predictions.
            nt_mass: 2D FloatTensor BS x T with the cumulated mass assigned to all number tokens.
        """
        self._validate_inputs(logits, labels=None, loss_weights=None)

        # Calculate the token-level predictions
        yhat = self._get_dot_product(logits=logits)

        probs_all = F.softmax(logits, dim=-1)
        probs_nt = probs_all[:, :, self.is_number_token]
        nt_mass = probs_nt.sum(dim=-1)
        return yhat, cast(FloatTensor, nt_mass)

    def _get_dot_product(
        self, logits: FloatTensor, number_token_positions: Optional[BoolTensor] = None
    ) -> FloatTensor:
        """
        Applies dot product of number token values and their predicted probabilites.

        Args:
            logits: 3D FloatTensor of shape BS x T x V.
            number_token_positions: Optional 2D BoolTensor (BS x T) containing locations
                of number tokens.

        Returns:
            If `number_token_positions` is None, 2D FloatTensor of shape BS x T.
            Otherwise, 1D FloatTensor containing the predictions for the number tokens.
        """
        # apply softmax solely over the number token indices
        nt_logits = logits[:, :, self.is_number_token]
        softmax_probs = F.softmax(nt_logits, dim=-1)
        values = self.number_values_dense.to(device=logits.device, dtype=logits.dtype)

        # compute the weighted average of number tokens
        if number_token_positions is None:
            # Calculate for all tokens
            yhat = torch.sum(softmax_probs * values, dim=-1)
        else:
            # Calculate selectively where labels are number tokens
            yhat = torch.sum(softmax_probs[number_token_positions] * values, dim=-1)
        return cast(FloatTensor, yhat)

    def forward(
        self,
        logits: FloatTensor,
        labels: LongTensor,
        loss_weights: Optional[Tensor] = None,
        reduction: str = "mean",
        ignore_index: int = -100,
    ) -> Tensor:
        """
        Computes the NTL based on the dot product between token values and their probs.

        Args:
            logits: 3D Tensor of shape BS x T x V.
            labels: 2D Tensor of shape BS x T.
            loss_weights: 2D Optional tensor of BS x T with token-wise loss weights.
            reduction: Optional string specifying the reduction to apply to the
                output. Defaults to "mean", options are "mean", "sum", "none".
            ignore_index: The token ID to ignore in the labels. Defaults to -100.

        Returns:
            Loss tensor
                OD if reduction=="mean"|"sum"
                BS x T if reduction=="none"
        """
        self._validate_inputs(logits, labels, loss_weights)

        y, loss_weights = self._prepare_number_token_targets(
            labels, loss_weights, ignore_index
        )
        loss_weights = loss_weights.to(logits.dtype)
        number_token_positions = cast(BoolTensor, ~torch.isnan(y))

        # If no digit tokens in batch, or total of the relevant loss weights is zero, no need for upcoming calculations
        if not number_token_positions.any() or not loss_weights.any():
            if (reduction == "mean") | (reduction == "sum"):
                loss = torch.tensor(0, dtype=logits.dtype, device=labels.device)
            elif reduction == "none":
                loss = torch.zeros_like(
                    labels, dtype=logits.dtype, device=labels.device
                )
            else:
                raise ValueError(f"{reduction} is not a valid value for reduction")

            return loss

        yhat = self._get_dot_product(
            logits=logits, number_token_positions=number_token_positions
        )

        # Apply specified loss function to y and yhat
        loss = self.loss_function(yhat, y[number_token_positions], reduction="none")

        # If reweigh: compute weights for NTL based on logits
        if self.reweigh:
            loss = self.reweigh_fn(
                logits=logits, loss=loss, number_token_positions=number_token_positions
            )

        loss = self._apply_reduction(
            loss=loss,
            reduction=reduction,
            loss_weights=loss_weights,
            number_token_positions=number_token_positions,
            logits=logits,
        )

        return loss


class NTLoss(AbstractNTLoss):
    """Class for Wasserstein-based NTLoss. This is the default in the ICML paper."""

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        digit_level: bool = True,
        reweigh: bool = True,
        squash_factor: Optional[float] = None,
    ):
        """
        NTL constructor for the Wasserstein-based NTLoss.

        Args:
            tokenizer: Any HuggingFace tokenizer.
            digit_level: Whether to ensure only digits are considered number tokens,
                stabilizing training with NTL. Defaults to True. Used for most
                experiments in the ICML paper.
            reweigh: Whether to scale the NTL using the logit weight on
                number tokens. Defaults to True.
                NOTE: The ICML paper does *not* use this option which can lead to
                incorrect loss if most mass is placed outside of the number tokens.
            squash_factor: The optional squashing factor for the NTL. If provided,
                this number denotes the factor by which predicting the largest number
                token is worse than predicting the closest incorrect number token.
                E.g., with digit-level tokenization this factor is 9. Setting this
                to 1 will recover cross entropy. This argument is intended to handle
                irregular vocabs with large numerical token values.
        """
        super().__init__(
            tokenizer=tokenizer,
            digit_level=digit_level,
            reweigh=reweigh,
        )

        self.squash_factor = squash_factor
        self.setup_distance_lookup(squash_factor)

    def setup_distance_lookup(
        self,
        squash_factor: Optional[float] = None,
    ) -> None:
        """
        Set up a lookup table for the distances between the number tokens.
        Use squash_factor to control by what factor the farthest number token is worse than the closest, incorrect number token.
        If not squash_factor is not set: with 10 number tokens (0-9), the squashing factor is 9.
        NOTE: With a squashing factor of 1, this basically collapses to cross entropy.

        Args:
            squash_factor: The optional squashing factor used.
        """

        # Get token ids for number tokens
        num_ids = torch.nonzero(self.is_number_token, as_tuple=True)[0]
        # Create mapping from number token ids to their index in order of appearance in vocab:
        # e.g. token "3" -> id 519 -> dist_idx 1, then abs dist to 3 for other NT values will be found in row/column 1
        vocab_to_dist_idx = torch.full((len(self.tokenizer),), -1, dtype=torch.long)
        # Use arange to ensure order of appearance
        vocab_to_dist_idx[num_ids] = torch.arange(num_ids.size(0), dtype=torch.long)

        # Build NxN abs-diff matrix
        vals = self.number_values_dense.unsqueeze(0)  # (1 x N)
        diff = torch.abs(vals - vals.t())  # (N x N)

        if isinstance(squash_factor, Number):
            assert squash_factor > 1, (
                f"The squash factor can't be equal to or smaller than 1, please use a different squashing factor than {squash_factor}"
            )

            # Mask out zeros to find the smallest nonzero diff
            inf = torch.finfo(diff.dtype).max
            diff_nonzero = diff.masked_fill(diff == 0, inf)
            global_min_nz = diff_nonzero.min()
            # Find largest diff
            global_max = diff.max()

            # Compute scaling factor based on indicated squash factor
            scale = (squash_factor - 1) / (global_max - global_min_nz)
            # Scale the absolute differences using scaling factor
            lookup = 1 + (diff - global_min_nz) * scale
            lookup[diff == 0] = 0.0

        else:
            lookup = diff

        self.vocab_to_dist_idx = vocab_to_dist_idx
        self.dist_lookup = lookup
        self.max_dist = lookup.max()

    def forward(
        self,
        logits: FloatTensor,
        labels: LongTensor,
        loss_weights: Optional[Tensor] = None,
        reduction: str = "mean",
        ignore_index: int = -100,
    ) -> Tensor:
        """
        Computes the NTL.

        Args:
            logits: 3D Tensor of shape BS x T x V.
            labels: 2D Tensor of shape BS x T.
            loss_weights: Optional 2D tensor of BS x T with token-specific weights.
            reduction: Optional string specifying the reduction to apply to the
                output. Defaults to "mean", options are "mean", "sum", "none".
            ignore_index: The token ID to ignore in the labels. Defaults to -100.

        Returns:
            Loss tensor
                OD if reduction=="mean"|"sum"
                BS x T if reduction=="none"

        """
        self._validate_inputs(logits, labels, loss_weights)

        y, loss_weights = self._prepare_number_token_targets(
            labels, loss_weights, ignore_index
        )
        loss_weights = loss_weights.to(logits.dtype)
        number_token_positions = ~torch.isnan(y)

        # If no digit tokens in batch, or total of the relevant loss_weights is zero, no need for upcoming calculations
        if not number_token_positions.any() or not loss_weights.any():
            if (reduction == "mean") | (reduction == "sum"):
                loss = torch.tensor(0, dtype=logits.dtype, device=labels.device)
            elif reduction == "none":
                loss = torch.zeros_like(
                    labels, dtype=logits.dtype, device=labels.device
                )
            else:
                raise ValueError(f"{reduction} is not a valid value for reduction")

            return loss

        # apply softmax and get number labels
        nt_logits = logits[:, :, self.is_number_token]
        softmax_probs = F.softmax(nt_logits, dim=-1)

        # get distance between the true numbers and all possible number values from lookup table
        abs_diff = self.dist_lookup.to(dtype=logits.dtype, device=logits.device)[
            self.vocab_to_dist_idx.to(device=labels.device)[
                labels[number_token_positions]
            ]
        ]

        # loss is the absolute difference weighted by the softmax probs
        loss = (abs_diff * softmax_probs[number_token_positions]).sum(dim=-1)

        # If reweigh: compute weights for NTL based on logits
        if self.reweigh:
            loss = self.reweigh_fn(
                logits=logits, loss=loss, number_token_positions=number_token_positions
            )

        loss = self._apply_reduction(
            loss=loss,
            reduction=reduction,
            loss_weights=loss_weights,
            number_token_positions=number_token_positions,
            logits=logits,
        )

        return loss


class NumberLevelLoss(NTLossDotProduct):
    """Class to calculate NTL on a per-number (rather than per-token) basis."""

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer,
        float_level: bool = False,
        reweigh: bool = True,
    ):
        """
        NTL constructor for the number-level NTLoss.

        Args:
            tokenizer: Any HuggingFace tokenizer.
            float_level: Whether to calculate the loss for every float or every
                integer in the sequence. For `12.34`, if float_level=False, two
                loss terms will be calculated, respectively for `12` and `34`.
                If float_level=True, a single `.` does not break the contiguity
                of the identified number. Defaults to False.
            reweigh: Whether to scale the NTL using the logit weight on
                number tokens. Defaults to True.
                NOTE: The ICML paper does *not* use this option which can lead to
                incorrect loss if most mass is placed outside of the number tokens.
                Using this will explode the NL-NTL in the current implementation,
                so reweighing for the NL-NTL needs to be refined.

        """
        # digit_level must be set to True.
        super().__init__(
            tokenizer=tokenizer,
            digit_level=True,
            reweigh=reweigh,
            loss_function=F.l1_loss,  # unused
        )
        self.float_level = float_level
        self.dot = self.tokenizer.convert_tokens_to_ids(".")

    def setup_max_dist(self):
        """
        Due to the MAPE loss calculation, the max dist is limited to 1.0
        """
        self.max_dist = torch.tensor(1.0)

    def convert_digits_to_numbers(
        self,
        y: FloatTensor,
        yhat: FloatTensor,
        number_token_positions: BoolTensor,
        labels: LongTensor,
    ):
        """
        Set up the order mask for the batch and convert digit-level number tokens to numerical values.

        Args:
            y: 2D FloatTensor of shape BS x T with target numerical values at digit-level (NaN for non-number tokens).
            yhat: 2D FloatTensor of shape BS x T containing the predictions for the number tokens at digit-level
                (includes predictions for non-number tokens).
            number_token_positions: 2D BoolTensor (BS x T) containing locations of number tokens at digit-level.
            labels: 2D LongTensor of shape BS x T with the target input IDs.

        Returns:
            y: 2D FloatTensor of shape BS x T with target numerical values at number-level (NaN for non-number tokens).
            yhat: 2D FloatTensor of shape BS x T containing the predictions for the number tokens at number-level
                (includes predictions for non-number tokens).
            number_token_positions: 2D BoolTensor (BS x T) containing locations of numerical values in y and yhat.
        """

        # Set up empty order_mask: will store power with which to scale digits
        order_mask = torch.zeros_like(y, dtype=yhat.dtype, device=y.device)

        # Extract numbers using number blocks
        for i in range(y.shape[0]):
            # For every item in batch: assume not starting with number block
            in_number_block = False
            end_digit = -1

            # Loop from end of sequence to beginning to extract numbers
            for j in range(y.shape[1] - 1, -1, -1):
                # Already in number block and a digit: increase order magnitude
                if in_number_block and number_token_positions[i, j]:
                    if not self.float_level or labels[i, j + 1] != self.dot:
                        previous_order_index = j + 1
                    else:
                        previous_order_index = j + 2
                    order_mask[i, j] = order_mask[i, previous_order_index] + 1

                # Not in number block: first instance of number = end digit
                elif number_token_positions[i, j]:
                    in_number_block = True
                    end_digit = j + 1

                # A dot can be considered part of a number if self.float_level
                elif (
                    in_number_block
                    and self.float_level
                    and labels[i, j] == self.dot
                    and labels[i, j + 1] != self.dot
                ):
                    # exp(-inf) = 0, thus, the dot does not contribute to the GT number calculation
                    order_mask[i, j] = -torch.inf
                    # Necessary to avoid having NaN when summing
                    y[i, j] = 0
                    yhat[i, j] = 0

                # In number block, but not a digit: end of number_block
                elif in_number_block:
                    in_number_block = False

                    # Reuse y and yhat tensors to store full numbers
                    y[i, j + 1] = torch.sum(
                        y[i, j + 1 : end_digit]
                        * torch.pow(10, order_mask[i, j + 1 : end_digit])
                    )
                    # Make sure non-relevant numerical values are turned into NaN
                    # This indicates non-number tokens
                    y[i, j + 2 : end_digit] = y[i, j]
                    yhat[i, j + 1] = torch.sum(
                        yhat[i, j + 1 : end_digit]
                        * torch.pow(10, order_mask[i, j + 1 : end_digit])
                    )

        # Update mask with locations of number tokens
        number_token_positions = cast(BoolTensor, ~torch.isnan(y))

        return y, yhat, number_token_positions

    def forward(
        self,
        logits: FloatTensor,
        labels: LongTensor,
        loss_weights: Optional[Tensor] = None,
        reduction: str = "mean",
        ignore_index: int = -100,
    ) -> Tensor:
        """
        Computes the NTL based on the dot product between token values and their probs.

        Args:
            logits: 3D Tensor of shape BS x T x V.
            labels: 2D Tensor of shape BS x T.
            loss_weights: 2D Optional tensor of BS x T with token-wise loss weights.
            reduction: Optional string specifying the reduction to apply to the
                output. Defaults to "mean", options are "mean", "sum", "none".
            ignore_index: The token ID to ignore in the labels. Defaults to -100.

        Returns:
            Loss tensor
                0-D if reduction=="mean"|"sum"
                BS x T if reduction=="none"
        """
        self._validate_inputs(logits, labels, loss_weights)

        y, _ = self._prepare_number_token_targets(labels, loss_weights, ignore_index)
        number_token_positions = cast(BoolTensor, ~torch.isnan(y))

        # If no digit tokens in batch, or total of the relevant loss weights is zero, no need for upcoming calculations
        if not number_token_positions.any() or (
            loss_weights is not None and not loss_weights.any()
        ):
            if (reduction == "mean") | (reduction == "sum"):
                loss = torch.tensor(0, dtype=logits.dtype, device=labels.device)
            elif reduction == "none":
                loss = torch.zeros_like(labels, dtype=logits.dtype)
            else:
                raise ValueError(f"{reduction} is not a valid value for reduction")

            return loss

        yhat = self._get_dot_product(logits=logits)

        y, yhat, number_token_positions = self.convert_digits_to_numbers(
            y, yhat, number_token_positions, labels
        )
        if loss_weights is None:
            loss_weights = torch.ones_like(labels, dtype=logits.dtype)
        loss_weights = loss_weights[number_token_positions]

        # NOTE: Alternative could be to apply specified loss function to normalized yhat
        # loss = self.loss_function(torch.div(
        #     yhat[number_token_positions],
        #     y[number_token_positions].clamp_min(torch.finfo(y.dtype).eps),
        # ), torch.ones_like(yhat), reduction="none")

        y_num = y[number_token_positions]
        yh_num = yhat[number_token_positions]
        # Calculate symmetric MAPE which is bounded in [0, 1]
        loss = (yh_num - y_num).abs() / (
            yh_num.abs() + y_num.abs() + torch.finfo(y.dtype).eps
        )

        # If reweigh: compute weights for NTL based on logits
        if self.reweigh:
            loss = self.reweigh_fn(
                logits=logits, loss=loss, number_token_positions=number_token_positions
            )

        if reduction == "mean":
            # Mean pooling (weighted by loss mask)
            loss = torch.dot(
                loss.flatten(), loss_weights.flatten()
            ) / loss_weights.sum().clamp_min(torch.finfo(loss.dtype).eps)
        elif reduction == "sum":
            loss = torch.dot(loss.flatten(), loss_weights.flatten())
        elif reduction == "none":
            # Cast loss for number tokens back to Tensor of size BS x T
            loss_ = torch.zeros(number_token_positions.numel()).to(loss.device)
            loss_[number_token_positions.view(-1)] = loss * loss_weights
            bs, seq_len, _ = logits.size()
            loss = loss_.view(bs, seq_len)

            assert torch.sum(loss[~number_token_positions]) == 0, (
                "NumberTokenLoss computed for non-digit tokens!"
            )

        else:
            raise ValueError(f"{reduction} is not a valid value for reduction")
        return loss
